// Programmer: Maura Dailey
// CSE 120 Section #: 1
// Date: 03/25/02

// Brief: This file defines the public member functions of the
//   SortedType class as set forth in the SortedType header file

#include <iostream>
#include <cstddef>

// SortedType Constructor: initializes length of the list
// to zero and sets the beginning of the list at listData
// to NULL

template <class ItemType>
SortedType<ItemType>::SortedType()
{
  length = 0;
  listData = NULL;
}

// SortedType Destructor: calls MakeEmpty() function to
// delete each node of the list, then deletes the first node
// of the list.

template <class ItemType>
SortedType<ItemType>::~SortedType()
{
  MakeEmpty();
  delete listData;
}

// IsFull(): attempts to create a new node in memory. If
// it fails, it returns that the list is full, otherwise
// it returns that the list is not yet full.

template <class ItemType>
bool SortedType<ItemType>::IsFull() const
{
  NodeType<ItemType>* ptr;

  ptr = new NodeType<ItemType>;
  if (ptr == NULL)
    return true;
  else
  {
    delete ptr;
    return false;
  }
}

// LengthIs(): returns the length of the list.

template <class ItemType>
int SortedType<ItemType>::LengthIs() const
{
  return length;
}

// MakeEmpty(): deletes each individual node of the list until
// listData equals NULL again, resetting the length to zero.

template <class ItemType>
void SortedType<ItemType>::MakeEmpty()
{
  // Follow the next pointers until listData equals NULL,
  // signifying that the list has been emptied
  while (listData != NULL)
  {
    NodeType<ItemType>* tempPtr = listData;
    listData = listData->next;
    delete tempPtr;
  }

  // Reset the length of the list to zero
  length = 0;
}

// DeleteItem(): Searches for item, reassigns the next and
// back pointers of the previous and next nodes to circumvent
// the node to be deleted, then deletes the original node
// that matched item.

template <class ItemType>
void SortedType<ItemType>::DeleteItem(ItemType item)
{
  bool moreToSearch = true;
  bool found = false;

  NodeType<ItemType>* location;

  // Provide an extra test to ensure the user is not attempting
  // to delete an item from an empty list, which is forbidden
  // in the Pre Conditions for the function.

  if (listData == NULL)
    return;
  else
    location = listData;

  // Search through the list until the item has been found.

  while (moreToSearch && !found)
  {
    if (location->info < item)
    {
      location = location->next;
      moreToSearch = (location != NULL);
    }
    else if (item == location->info)
    {
      found = true;
      moreToSearch = false;
    }
    else
    {
      moreToSearch = false;
    }
  }

  // If the search does not find the item, a condition forbidden
  // by the Pre Conditions for the function, exit the function.
  if (found == false)
    return;

  NodeType<ItemType>* tempPtr;

  if (location != NULL)
  {
    // If the list has a length of one, then it is safe to simply
    // call the MakeEmpty() function.
    if (LengthIs() == 1)
      MakeEmpty();

    // If the item is the last item in the list, set the forward
    // pointer of the previous node to be NULL. Delete the item.
    else if (location->next == NULL)
    {
      tempPtr = location;
      (location->back)->next = NULL;
      delete tempPtr;
      length--;
    }

    // If the item is at the beginning of the list, set the pointer
    // to the beginning of the list to be the next node, while
    // setting the back pointer to be NULL. Delete the item.
    else if (location->back == NULL)
    {
      tempPtr = location;
      listData = location->next;
      (location->next)->back = NULL;
      delete tempPtr;
      length--;
    }

    // If the item is somewhere inside the list, set the back
    // pointer of the next node in the list to point to the
    // previous node, and set the next pointer of the previous
    // node in the list to point to the next node. Delete the item.
    else	
    {
      tempPtr = location;
      (location->back)->next = location->next;
      (location->next)->back = location->back;
      delete tempPtr;
      length--;
    }
  }
}

// RetrieveItem(): Search the list until item has been found
// or it is discovered that item is not in the list. Set the
// boolean variable found to true if the item is in the list,
// else set to false.

template <class ItemType>
void SortedType<ItemType>::RetrieveItem(ItemType& item, bool& found)
{
  bool moreToSearch = false;
  NodeType<ItemType>* location;

  location = listData;
  found = false;
  
  // As long as the list is not empty, go ahead and search for item
  moreToSearch = (location != NULL);

  // Search until item has been found or the location where item
  // should be contained has been passed
  while (moreToSearch && !found)
  {
    // If the item is greater than the value of the current node,
    // advance through the list another node.
    if (location->info < item)
    {
      location = location->next;
      moreToSearch = (location != NULL);
    }

    // If the item is equal to the value of the current node,
    // set found to true and set item to be the value of the
    // current node.
    else if (item == location->info)
    {
      found = true;
      item = location->info;
      moreToSearch = false;
    }
    
    // If the item is not in the list, simply exit the search
    // and end the function.
    else
      moreToSearch = false;
  }
}

// InsertItem(): searches the list for the proper place to insert
// a new node such that the order of the list remains intact.
// Changes the front and back pointers of the back and next
// nodes respectively to make room for the new node.

template <class ItemType>
void SortedType<ItemType>::InsertItem(ItemType item)
{
  // Make sure the list is not full.
  if (IsFull())
    return;

  NodeType<ItemType>* location;
  NodeType<ItemType>* newNode;
  NodeType<ItemType>* prevLoc;

  location = listData;
  prevLoc = NULL;

  bool moreToSearch = (location != NULL);

  // Search the list to find the proper place to insert item
  // in order to maintain the sorted order.
  while (moreToSearch && location != NULL)
  {
    if (location->info < item)
    {
      prevLoc = location;
      location = location->next;
      moreToSearch = (location != NULL);
    }
    else
      moreToSearch = false;
  }

  // Allocate space for the newNode, and set the current
  // value of the node to item.
  newNode = new NodeType<ItemType>;
  newNode->info = item;

  // If the list is empty, set the back and next pointers
  // of the newNode to point to NULL. Set the newNode to be
  // the beginning of the list.
  if (location == NULL & prevLoc == NULL)
  {
    newNode->back = NULL;
    newNode->next = NULL;
    listData = newNode;
  }

  // If the newNode will be inserted at the end of the list,
  // set the next pointer of newNode to be NULL, the back
  // pointer to be the previous node, and the next pointer
  // of the previous node to point to the newNode.
  else if (location == NULL & prevLoc != NULL)
  {
    prevLoc->next = newNode;
    newNode->back = prevLoc;
    newNode->next = NULL;
  }

  // If the newNode will be inserted at the beginning of the
  // list, set the back pointer of the newNode to be NULL, the
  // next pointer of newNode to point to the next node in the
  // list, and the back pointer of the next node to point to
  // the newNode. Set the newNode to be the beginning of the
  // list.
  else if (location->back == NULL)
  {
    newNode->next = listData;
    newNode->back = NULL;
    location->back = newNode;
    listData = newNode;
  }

  // If the newNode will be inserted somewhere in the middle
  // of the list, set the next and back pointers of newNode to
  // point to the next and previous nodes respectively. Set the
  // back and next pointers of those nodes to point to newNode.
  else
  {
    newNode->back = location->back;
    newNode->next = location;
    (location->back)->next = newNode;
    location->back = newNode;
  }

  // Increment length to acknowledge the addition of a node to
  // the list.
  length ++;
}

// ResetList(): resets the current position to be at the beginning
// of the list.

template <class ItemType>
void SortedType<ItemType>::ResetList()
{
  currentPos = NULL;
}

// GetPrevItem(): provided that the list is not empty and that the
// current position is not at the first item in the list, item will
// be set to the value of the previous node and found set to true.
// Else, found is set to false.

template <class ItemType>
void SortedType<ItemType>::GetPrevItem(ItemType& item, bool& found)
{
  // Test for empty list
  if (length == 0)
    found = false;

  else
  {
    // If currentPos is NULL, then it must be at the beginning of the
    // list such that there are no previous items to be retrieved
    if (currentPos == NULL)
      found = false;

    // If the node to the left of currentPos is NULL, then the previous
    // item should be listData, the first item in the list.
    else if (currentPos->back == NULL)
    {
      item = listData->info;
      found = true;
    }

    // Otherwise, the currentPos is already at the previous item, so
    // it is safe to retrieve the item. 
    else
    {
      item = currentPos->info;
      currentPos = currentPos->back;
      found = true;
    }
  }
}

// GetNextItem(): Provided that the list is not empty and that the
// current position is not at the last item in the list, item will
// be set to the value of the next node and found set to true. Else,
// found will be set to false.

template <class ItemType>
void SortedType<ItemType>::GetNextItem(ItemType& item, bool& found)
{
  // Test for empty list
  if (length == 0)
    found = false;

  else
  {
    // If currentPos is NULL, then the next item will be listData, the
    // first item in the list.
    if (currentPos == NULL)
    {
      currentPos = listData;
      item = currentPos->info;
      found = true;
    }

    // If the node to the right of currentPos is not NULL, then the list
    // is not at the end and the next item can be retrieved.
    else if (currentPos->next != NULL)
    {
      currentPos = currentPos->next;
      item = currentPos->info;
      found = true;
    }

    // Otherwise, currentPos must be on the last item in the list, so
    // the next item can not be retrieved.
    else
    {
      found = false;
    }
  }
}

// SortedType(): Overloaded constructor needed for the overloaded operator
//   that follows.

template <class ItemType>
SortedType<ItemType>::SortedType(const SortedType<ItemType>& newList)
{
  // Test to see if the list being copied is empty and if so, make the
  // new list empty as well.
  if (newList.listData == NULL)
    listData=NULL;

  else
  {	
    NodeType<ItemType>* location;
    NodeType<ItemType>* nextLocation;

    // Create the first node of the list, listData
    listData = new NodeType<ItemType>;
    listData->info = (newList.listData)->info;
    listData->back = NULL;
    nextLocation = (newList.listData)->next;
    location = listData;

    // Until the last item has been reached in the original list,
    // continue to copy each node
    while (nextLocation != NULL)
    {
      location->next = new NodeType<ItemType>;
      (location->next)->back = location;
      location = location->next;
      location->info = nextLocation->info;
      nextLocation = nextLocation->next;
    }

    location->next = NULL;
  }
}

// operator=(): makes an exact copy of what comes on the right side of
// the equals sign and assigns it to a new list variable on the left side
// of the equals sign.

template <class ItemType>
void SortedType<ItemType>::operator=(SortedType<ItemType> newList)
{
  MakeEmpty();

  // Test to see if the list being copied is empty and if so, make the
  // new list empty as well.
  if (newList.listData == NULL)
    listData = NULL;

  else
  {	
    NodeType<ItemType>* location;
    NodeType<ItemType>* nextLocation;

    // Create the first node of the list, listData
    listData = new NodeType<ItemType>;
    listData->info = (newList.listData)->info;
    listData->back = NULL;
    nextLocation = (newList.listData)->next;
    location = listData;

    // The length of this list is now one
    length = 1;

    // Until the last item has been reached in the original list,
    // continue to copy each node
    while (nextLocation != NULL)
    {
      location->next = new NodeType<ItemType>;
      (location->next)->back = location;
      location = location->next;
      location->info = nextLocation->info;
      nextLocation = nextLocation->next;
      length++;
    }

    location->next = NULL;
  }
}